/*
 * Copyright (C) 2014-2016 Brian L. Browning
 *
 * This file is part of Refined-IBD
 *
 * Refined-IBD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Beagle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package ibd;

import beagleutil.IntInterval;
import blbutil.Const;

/**
 * <p>Class {@code HapPairSegment} represents a marker interval
 * for a haplotype.
 * </p>
 * Instances of class {@code HapPairSegment} are immutable.
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public class HapPairSegment implements Comparable<HapPairSegment>, IntInterval {

    private final int hap1;
    private final int hap2;
    private final int start;
    private final int end;

    /**
     * Constructs a new {@code HapPairSegment} instance.
     * @param hap1 the first haplotype index
     * @param hap2 the second haplotype index
     * @param start the start marker index (inclusive)
     * @param end the end marker index (inclusive)
     * @throws IllegalArgumentException if {@code start > end}
     */
    public HapPairSegment(int hap1, int hap2, int start, int end) {
        if (start > end) {
            throw new IllegalArgumentException(String.valueOf(start));
        }
        this.hap1 = hap1;
        this.hap2 = hap2;
        this.start = start;
        this.end = end;
    }

    /**
     * Returns the first haplotype index.
     * @return the first haplotype index
     */
    public int hap1() {
        return hap1;
    }

    /**
     * Returns the second haplotype index.
     * @return the second haplotype index
     */
    public int hap2() {
        return hap2;
    }

    /**
     * Returns the start marker index (inclusive).
     * @return the start marker index (inclusive)
     */
    @Override
    public int start() {
        return start;
    }

    /**
     * Returns the end marker index (inclusive).
     * @return the end marker index (inclusive)
     */
    @Override
    public int end() {
        return end;
    }

    /**
     * Returns a string representation of {@code this}.  The exact
     * details of the representation are unspecified and subject to change.
     * @return a string representation of {@code this}
     */
    @Override
    public String toString() {
        StringBuilder sb = new StringBuilder(70);
        sb.append(hap1);
        sb.append(Const.tab);
        sb.append(hap2);
        sb.append(Const.tab);
        sb.append(start);
        sb.append(Const.tab);
        sb.append(end);
        return sb.toString();
    }

    /**
     * <p>Returns the hash code value for this object. The hash code is defined
     * by the following calculation:
     * </p>
     * <pre>
     *  int hash = 5;
     *  hash = 89 * hash + this.hap1();
     *  hash = 89 * hash + this.hap2();
     *  hash = 89 * hash + this.start();
     *  hash = 89 * hash + this.end();
     </pre>
     * @return the hash code value for this object
     */
    @Override
    public int hashCode() {
        int hash = 5;
        hash = 89*hash + this.hap1;
        hash = 89*hash + this.hap2;
        hash = 89*hash + this.start;
        hash = 89*hash + this.end;
        return hash;
    }

    /**
     * Compares the specified object with this {@code HapPairSegment} for
     * equality. Returns {@code true} if the specified object is a
     * {@code HapPairSegment} instance and if this {@code HapPairSegment} is
     * equal to the specified {@code HapPairSegment}, and returns
     * {@code false}  otherwise.  Two {@code HapPairSegment}  instances
     * are equal if they have equal haplotype indices,
     * equal starting marker indices, and equal ending marker indices.
     * @param o the reference object with which to compare.
     * @return {@code true} if the specified object is an
     * {@code HapPairSegment} instance and if this {@code HapPairSegment} is
     * equal to the specified {@code HapPairSegment}
     */
    @Override
    public boolean equals(Object o) {
        if (o==null) {
            return false;
        }
        if (getClass()!=o.getClass()) {
            return false;
        }
        final HapPairSegment other=(HapPairSegment) o;
        return (this.hap1==other.hap1 && this.hap2==other.hap2
                && this.start==other.start && this.end==other.end);
    }

    /**
     * Compares this object with the specified object for order.  Returns a
     * negative integer, zero, or a positive integer as this object is less
     * than, equal to, or greater than the specified object.
     * {@code HapPairSegment} instances are ordered first by
     * {@code this.start()}, then by {@code this.end()},
     * then by {@code this.hap1()}, and finally by {@code this.hap2()}.
     * @param hs the {@code HapPairSegment} to be compared
     * @return a negative integer, zero, or a positive integer as this
     * {@code HapPairSegment} is less than, equal to, or greater than the
     * specified {@code HapPairSegment}
     * @throws NullPointerException if {@code o == null}
     */
    @Override
    public int compareTo(HapPairSegment hs) {
        if (this.start != hs.start) {
            return (this.start < hs.start) ? -1 : 1;
        }
        else if (this.end != hs.end) {
            return (this.end < hs.end) ? -1 : 1;
        }
        if (this.hap1 != hs.hap1) {
            return (this.hap1 < hs.hap1) ? -1 : 1;
        }
        if (this.hap2 != hs.hap2) {
            return (this.hap2 < hs.hap2) ? -1 : 1;
        }
        return 0;
    }
}
