/*
 * Copyright (C) 2014-2016 Brian L. Browning
 *
 * This file is part of Refined-IBD
 *
 * Refined-IBD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Beagle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package refined;

import blbutil.IntPair;
import blbutil.Utilities;
import dag.Dag;
import ibd.BeagleIbdSeg;
import ibd.HapPairSegment;
import ibd.IbsHapSegments;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ArrayBlockingQueue;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.TimeUnit;
import vcf.GT;
import vcf.GeneticMap;
import vcf.PhasedGT;

/**
 * <p>Class {@code HaploidIbd} implements the Refined IBD algorithm.
 * The Refined IBD algorithm detects candidate haplotype IBD segments with the
 * Germline Algorithm and then evaluates candidate IBD segments using a
 * likelihood ratio test.
 * </p>
 * <p>Instances of class {@code HaploidIbd} are immutable.
 *</p>
 * Reference: Gusev A, Lowe JK, Stoffel M, Daly MJ, Altshuler D, Breslow JL,
 *      Friedman JM, Pe'er I.  Whole population, genomewide mapping
 *      of hidden relatedness.  Genome Research 2009;19(2):318-26.
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public final class HaploidIbd {

    private final GeneticMap genMap;
    private final float ibdTrim;
    private final float minIbdLod;
    private final float minFreqLod;
    private final float minCm;

    /**
     * Constructs a new {@code HaploidIbd} instance from the specified data.
     * @param genMap the genetic map
     * @param ibdTrim the number of markers to trim from an IBS segment
     * when computing the IBD versus non-IBD likelihood ratio
     * @param minIbdLod the minimum IBD LOD score of reported IBD segments
     * @param minCm the minimum cM length of reported IBD segments
     *
     * @throws IllegalArgumentException if {@code ibdTrim < 0 }
     * @throws IllegalArgumentException if
     * {@code ibdLod <= 0.0f || Float.isFinite(ibdLod) == false}
     * @throws IllegalArgumentException if
     * {@code minCm <= 0.0f || Float.isFinite(minCm) == false}
     * @throws NullPointerException if {@code genMap == null}
     */
    public HaploidIbd(GeneticMap genMap, float ibdTrim, float minIbdLod,
            float minCm) {
        if (genMap==null) {
            throw new IllegalArgumentException(GeneticMap.class.toString());
        }
        if (ibdTrim < 0f) {
            throw new IllegalArgumentException(String.valueOf(ibdTrim));
        }
        if (minIbdLod <= 0f || Float.isFinite(minIbdLod) == false) {
            throw new IllegalArgumentException(String.valueOf(minIbdLod));
        }
        if (minCm <= 0f || Float.isFinite(minCm) == false) {
            throw new IllegalArgumentException(String.valueOf(minCm));
        }
        this.genMap = genMap;
        this.ibdTrim = ibdTrim;
        this.minIbdLod = minIbdLod;
        this.minFreqLod = minIbdLod;
        this.minCm = minCm;
    }

    /**
     * Runs the Refined IBD algorithm, and returns a map whose keys are
     * ordered pairs of haplotype indices and whose values are thread-safe
     * lists of IBD segments for each haplotype pair. The minimum haplotype
     * index is listed first in each ordered pair of haplotype indices.
     *
     * @param gl the HMM emission probabilities
     * @param dag the HMM transition probabilities
     * @param haps the sample haplotype pairs
     * @param nThreads the number of threads of execution that may be used
     * @return the detected IBD segments
     *
     * @throws IllegalArgumentException if {@code nThreads < 1}
     * @throws IllegalArgumentException if
     * {@code gl.samples().equals(hapPairs.samples()) == false}
     * @throws IllegalArgumentException if
     * {@code gl.markers().equals(dag.markers()) == false
                || gl.markers().equals(hapPairs.markers()) == false}
     * @throws NullPointerException if
     * {@code gl == null || dag == null || hapPairs == null}
     */
    @SuppressWarnings({"BroadCatchBlock", "TooBroadCatch"})
    public Map<IntPair, List<BeagleIbdSeg>> run(GT gl, Dag dag,
            PhasedGT haps, final int nThreads) {
        checkParameters(gl, dag, haps);
        double[] pos = genMap.genPos(dag.markers());
        IbsHapSegments ibsSegments = new IbsHapSegments(haps, pos, minCm);
        ConcurrentMap<IntPair, List<BeagleIbdSeg>> ibdMap
                = new ConcurrentHashMap<>();

        final BlockingQueue<Integer> qIn = new ArrayBlockingQueue<>(5*nThreads);
        ExecutorService es = Executors.newFixedThreadPool(nThreads);
        for (int j=0; j<nThreads; ++j) {
            IbdBaum baum = new IbdBaum(dag, gl);
            es.submit(new ProduceIbd(haps, pos, baum, ibsSegments, qIn, ibdMap,
                    ibdTrim, minIbdLod));
        }
        try {
            for (int hap=0, n=haps.nHaps(); hap<n; ++hap) {
                qIn.put(hap);
            }
            for (int j=0; j<nThreads; ++j) {
               qIn.put(ProduceIbd.POISON);
            }
            es.shutdown();
            es.awaitTermination(Long.MAX_VALUE, TimeUnit.DAYS);
        }
        catch (Throwable e) {
            Utilities.exit("ERROR", e);
        }
        return ibdMap;
    }

    private void checkParameters(GT gl, Dag dag, PhasedGT haps) {
        if (gl.samples().equals(haps.samples())==false) {
            throw new IllegalArgumentException("inconstent samples");
        }
        if (gl.markers().equals(dag.markers())==false
                || gl.markers().equals(haps.markers())==false) {
            throw new IllegalArgumentException("inconsistent markers");
        }
    }

    private static double freqLod(int hap, int start, int end,  Dag dag,
            PhasedGT haps) {
        return start>=end ? 0.0f : IbdBaum.freqLod(hap, start, end+1, haps, dag);
    }

    private static double ibdLod(IbdBaum ibdBaum, int hap1, int hap2,
            int start, int end) {
        return start>=end ? 0.0f : ibdBaum.ibdLod(hap1/2, hap2/2, start, end+1);
    }

    private class ProduceIbd implements Runnable {

        public static final int POISON = -37;

        private final PhasedGT haps;
        private final double[] pos;
        private final IbdBaum baum;
        private final IbsHapSegments ibsHapSegments;
        private final BlockingQueue<Integer> qIn;
        private final ConcurrentMap<IntPair, List<BeagleIbdSeg>> ibdMap;

        public ProduceIbd(PhasedGT haps, double[] pos, IbdBaum baum,
                IbsHapSegments ibsHapSegments, BlockingQueue<Integer> qIn,
                ConcurrentMap<IntPair, List<BeagleIbdSeg>> ibdMap, float ibdTrim,
                float minIbdLod) {
            if (ibdTrim < 0) {
                throw new IllegalArgumentException("trim < 0: " + ibdTrim);
            }
            if (minIbdLod <= 0.0 || Float.isNaN(minIbdLod)) {
                throw new IllegalArgumentException("ibdlod: " + minIbdLod);
            }
            this.haps = haps;
            this.pos = pos;
            this.baum = baum;
            this.ibsHapSegments = ibsHapSegments;
            this.qIn = qIn;
            this.ibdMap = ibdMap;
        }

        /*
         * Takes haplotype indices from a thread-safe work-queue and stores
         * detected IBD segments that between the haplotype and
         * haplotypes with larger index in {@code this.ibdMap}.  The method
         * exits when {@code ProduceSingleSamples.POISON} is taken from the
         * work queue.
         *
         * @throws IndexOutOfBounds exception if a negative integer
         * other than {@code ProduceSingleSamples.POISON} is taken from the
         * work queue
         */
        @Override
        @SuppressWarnings({"BroadCatchBlock", "TooBroadCatch"})
        public void run() {
            try {
                int hap = qIn.take();
                while (hap!=POISON) {
                    List<HapPairSegment> ibsSegs = ibsHapSegments.find(hap);
                    for (int j=0, n=ibsSegs.size(); j<n; ++j) {
                        HapPairSegment hps = ibsSegs.get(j);
                        if (hps.hap1() < hps.hap2()) {
                            int start = hps.start();
                            int end = hps.end();
                            int trimmedStart = trimmedStart(start);
                            int trimmedEnd = trimmedEnd(end);
                            double freqLod = freqLod(hap, trimmedStart,
                                    trimmedEnd, baum.dag(), haps);
                            if (freqLod >= minFreqLod) {
                                float ibdLod;
                                if ( (hps.hap1()/2) == (hps.hap2()/2) ) {
                                    int sample = hap/2;
                                    // no trim for HBD because phase is known
                                    ibdLod = (float) baum.hbdLod(sample, start, (end+1));
                                }
                                else {
                                    ibdLod = (float) ibdLod(baum, hps.hap1(),
                                            hps.hap2(), trimmedStart, trimmedEnd);
                                }
                                if (ibdLod >= minIbdLod) {
                                    IntPair hapPair = new IntPair(hps.hap1(), hps.hap2());
                                    List<BeagleIbdSeg> list = ibdMap.get(hapPair);
                                    if (list==null) {
                                        list = Collections.synchronizedList(
                                                new ArrayList<BeagleIbdSeg>(2));
                                        ibdMap.putIfAbsent(hapPair, list);
                                        list = ibdMap.get(hapPair);
                                    }
                                    BeagleIbdSeg segment = new BeagleIbdSeg(hapPair,
                                            baum.gl().marker(start),
                                            baum.gl().marker(end),
                                            ibdLod, start, end );
                                    list.add(segment);
                                }
                            }
                        }
                    }
                    hap = qIn.take();
                }
            }
            catch (Throwable t) {
                Utilities.exit(t);
            }
        }
        private int trimmedStart(int start) {
            int index = Arrays.binarySearch(pos, pos[start] + ibdTrim);
            if (index>0) {
                while (index>0 && pos[index]==pos[index-1]) {
                    --index;
                }
                return index;
            }
            else {
                int insPt = -index - 1;
                if (insPt==pos.length) {
                    --insPt;
                }
                return insPt;
            }
        }

        private int trimmedEnd(int inclusiveEnd) {
            int index = Arrays.binarySearch(pos, pos[inclusiveEnd] - ibdTrim);
            if (index>=0) {
                while (index+1<pos.length && pos[index]==pos[index+1]) {
                    ++index;
                }
                return index;
            }
            else {
                int insPt = -index - 1;
                return insPt>0 ? (insPt - 1) : insPt;
            }
        }
    }
}
