/*
 * Copyright (C) 2014-2016 Brian L. Browning
 *
 * This file is part of Refined-IBD
 *
 * Refined-IBD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Beagle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package refined;

import beagleutil.ChromInterval;
import beagleutil.Samples;
import blbutil.Const;
import blbutil.FileIt;
import blbutil.FileUtil;
import blbutil.Filter;
import blbutil.InputIt;
import vcf.GeneticMap;
import blbutil.IntPair;
import blbutil.SampleFileIt;
import dag.Dag;
import dag.MergeableDag;
import haplotype.HapPairs;
import ibd.BeagleIbdSeg;
import haplotype.WrappedHapPairs;
import java.io.File;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.Arrays;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.function.Supplier;
import vcf.BasicRefGT;
import vcf.FilterUtil;
import vcf.GT;
import vcf.IntervalVcfIt;
import vcf.Marker;
import vcf.NoPhaseGTWrapper;
import vcf.PhasedGT;
import vcf.RefGT;
import vcf.RefGTRec;
import vcf.RefIt;
import vcf.Window;
import vcf.WindowIt;

/**
 * <p>Class {@code RefinedIbd} runs a refined-ibd analysis.
 * </p>
 * <p>Instances of class {@code RefinedIbd} are not thread-safe.
 * </p>
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public class RefinedIbd {

    private static final DecimalFormat DF2 = new DecimalFormat("0.##");
    private static final DecimalFormat DF3 = new DecimalFormat("0.###");

    private final RefinedPar par;
    private final Map<IntPair, BeagleIbdSeg> crossSpliceBuffer;

    /**
     * Constructs a new {@code MainHelper} instance.
     * @param par the command line parameters
     * @throws NullPointerException if {@code par == null}
     */
    public RefinedIbd(RefinedPar par) {
        if (par==null) {
            throw new NullPointerException(RefinedPar.class.toGenericString());
        }
        this.par = par;
        this.crossSpliceBuffer = new HashMap<>();
    }

    /**
     * Returns the refined-ibd analysis parameters.
     * @return the refined-ibd analysis parameters
     */
    public RefinedPar par() {
        return par;
    }

    /**
     * Performs a refined-ibd IBD segment detection analysis using
     * the analysis parameters from {@code this.par()}.
     */
    public void run() {
        int windowIndex = 0;
        try (WindowIt<RefGTRec> windowIt = windowIt(par);
                RefinedScreenOutput runStats = new RefinedScreenOutput(par)) {
            GeneticMap genMap = windowIt.genMap();
            runStats.printSampleSummary(windowIt.samples().nSamples());
            while (windowIt.hasNext()) {
                ++windowIndex;
                Window<RefGTRec> wind = windowIt.next();
                RefGTRec[] recs = wind.recList().toArray(new RefGTRec[0]);
                RefGT phasedGT = new BasicRefGT(recs);
                runStats.printWindow(phasedGT.markers(), windowIndex);

                long t0 = System.nanoTime();
                Dag dag = dag(phasedGT, par.adjustedIbdScale(phasedGT.nSamples()));
                long t1 = System.nanoTime();
                Map<IntPair, List<BeagleIbdSeg>> ibd = ibd(dag, phasedGT, genMap);
                long t2 = System.nanoTime();

                long dagNanos = (t1-t0);
                long ibdNanos = (t2-t1);

                runStats.printNMarkers(phasedGT.nMarkers());
                runStats.printTiming(dagNanos, ibdNanos, (dagNanos + ibdNanos));
                runStats.printDagStats(dag);
                runStats.addToMarkerCnt(wind.nMarkers() - wind.overlapEnd());

                boolean append = (windowIndex > 1);
                printIbd(windowIt.samples(), wind, genMap, ibd, append);
            }
        }
    }

    private static WindowIt<RefGTRec> windowIt(RefinedPar par) {
        Filter<String> sFilter = FilterUtil.sampleFilter(par.excludesamples());
        Filter<Marker> mFilter = FilterUtil.markerFilter(par.excludemarkers());
        ChromInterval chromInterval = par.chromInt();
        GeneticMap map = GeneticMap.geneticMap(par.map(), par.chromInt());

        Supplier<SampleFileIt<RefGTRec>> supplier = () -> {
            FileIt<String> it0 = InputIt.fromGzipFile(par.gt());
            SampleFileIt<RefGTRec> it1 = RefIt.create(it0, sFilter, mFilter,
                    RefIt.MAX_EM_BUFFER_SIZE);
            if (chromInterval!=null) {
                it1 = new IntervalVcfIt<>(it1, chromInterval);
            }
            return it1;
        } ;
        float overlap = 2.5f*par.length();  // since splice point may not be genetic mid-point.
        return WindowIt.newInstance(supplier, map, par.window(), overlap);
    }

    private void printIbd(Samples samples, Window<RefGTRec> window,
            GeneticMap genMap, Map<IntPair, List<BeagleIbdSeg>> ibd,
            boolean append) {
        int nextSplice = (window.overlapStart() + window.nMarkers()) >>> 1;
        int prevSplice = (window.overlapEnd() / 2);

        Map<IntPair, BeagleIbdSeg> lastBuffer = new HashMap<>(crossSpliceBuffer);
        crossSpliceBuffer.clear();
        File ibdFile = new File(par.out() + ".ibd.gz");
        File hbdFile = new File(par.out() + ".hbd.gz");
        try (PrintWriter ibdOut = FileUtil.gzipPrintWriter(ibdFile, append);
               PrintWriter hbdOut = FileUtil.gzipPrintWriter(hbdFile, append)) {
            Iterator<IntPair> keyIt = ibd.keySet().iterator();
            while (keyIt.hasNext()) {
                IntPair key = keyIt.next();
                List<BeagleIbdSeg> list = ibd.get(key);
                for (BeagleIbdSeg seg : list) {
                    // seg.endIndex() is inclusive end
                    if (seg.endIndex() >= prevSplice) { // not printed in previous window
                        if (seg.startIndex() < prevSplice) {
                            BeagleIbdSeg saved = lastBuffer.remove(key);
                            if (saved!=null) {
                                seg = merge(saved, seg);
                            }
                        }
                        if (seg.endIndex() < nextSplice) {
                            printSegment(samples, genMap, seg, ibdOut, hbdOut);
                        }
                        else if (seg.startIndex() < nextSplice) {
                            BeagleIbdSeg prev = crossSpliceBuffer.put(key, seg);
                            assert prev==null;
                        }
                    }
                }
                keyIt.remove();
            }
            printBuffer(samples, genMap, lastBuffer, ibdOut, hbdOut);
        }
    }

    private void printBuffer(Samples samples, GeneticMap genMap,
            Map<IntPair, BeagleIbdSeg> buffer, PrintWriter ibdOut,
            PrintWriter hbdOut) {
        Iterator<IntPair> keyIt = buffer.keySet().iterator();
        while (keyIt.hasNext()) {
            BeagleIbdSeg seg = buffer.get(keyIt.next());
            printSegment(samples, genMap, seg, ibdOut, hbdOut);
        }
    }

    private static BeagleIbdSeg merge(BeagleIbdSeg a, BeagleIbdSeg b) {
        assert a.hapPair().equals(b.hapPair());
        assert a.start().chromIndex()==b.start().chromIndex();
        int newStartIndex = -1;
        float newScore = Math.max(a.score(), b.score());
        return new BeagleIbdSeg(a.hapPair(), a.start(), b.end(),
                newScore, newStartIndex, b.endIndex());
    }

    private static void printSegment(Samples samples, GeneticMap genMap,
            BeagleIbdSeg tract, PrintWriter ibdOut, PrintWriter hbdOut) {
        int h1 = tract.hap1();
        int h2 = tract.hap2();
        int s1 = h1/2;
        double length = genMap.genPos(tract.end()) - genMap.genPos(tract.start());
        int s2 = h2/2;
        PrintWriter out = (s1==s2) ? hbdOut : ibdOut;
        out.print(samples.id(s1));
        out.print(Const.tab);
        out.print((h1 % 2) + 1);
        out.print(Const.tab);
        out.print(samples.id(s2));
        out.print(Const.tab);
        out.print((h2 % 2) + 1);
        out.print(Const.tab);
        out.print(tract.start().chrom());
        out.print(Const.tab);
        out.print(tract.start().pos());
        out.print(Const.tab);
        out.print(tract.end().pos());
        out.print(Const.tab);
        out.print(DF2.format(tract.score()));
        out.print(Const.tab);
        out.println(DF3.format(length));
    }

    private Map<IntPair, List<BeagleIbdSeg>> ibd(Dag dag, PhasedGT phasedGT,
            GeneticMap genMap) {
        GT noPhaseGT = new NoPhaseGTWrapper(phasedGT);
        HaploidIbd hapIbd = new HaploidIbd(genMap, par.trim(),
                par.lod(), par.length());
        Map<IntPair, List<BeagleIbdSeg>> ibdMap =
                hapIbd.run(noPhaseGT, dag, phasedGT, par.nthreads());
        return ibdMap;
    }

    private Dag dag(PhasedGT phasedTarg, float scale) {
        HapPairs dagHaps = new WrappedHapPairs(phasedTarg);
        float[] weights = new float[dagHaps.nHaps()];
        Arrays.fill(weights, 1.0f);
        int nInitLevels = 500;
        Dag ibdDag = MergeableDag.dag(dagHaps, weights, scale, nInitLevels);
        return ibdDag;
    }
}
