/*
 * Copyright (C) 2014-2016 Brian L. Browning
 *
 * This file is part of Refined-IBD
 *
 * Refined-IBD is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Beagle is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package refined;

import blbutil.Const;
import blbutil.Utilities;
import java.io.File;
import java.util.Locale;

/**
 * Class {@code RefinedMain} is the entry class for the refined-ibd program.
 * See {@code RefinedPar.usage()} and online program documentation for usage
 * instructions.
 *
 * @author Brian L. Browning {@code <browning@uw.edu>}
 */
public class RefinedMain {

    /**
     * The copyright string.
     */
    private static final String COPYRIGHT = "Copyright (C) 2014-2017 Brian L. Browning";

    static final String PROGRAM = "refined-ibd.17Jan20.102.jar";
    static final String COMMAND = "java -jar refined-ibd.17Jan20.102.jar";
    static final String SHORT_HELP = PROGRAM + Const.nl + COPYRIGHT
            + Const.nl + "Enter " + COMMAND + " to list command line argument";

    /**
     * <p>Entry point to the refined-ibd program. See {@code RefinedPar.usage()}
     * and program documentation for usage instructions.
     * </p>
     *
     * @param args command line arguments
     */
    public static void main(String[] args) {
	Locale.setDefault(Locale.US);
        if (args.length==0) {
            System.out.println(PROGRAM);
            System.out.println(COPYRIGHT);
            System.out.println(RefinedPar.usage());
            System.exit(0);
        }
        RefinedPar par = parameters(args);
        System.setProperty("java.util.concurrent.ForkJoinPool.common.parallelism",
                String.valueOf(par.nthreads()));

        RefinedIbd ribd = new RefinedIbd(par);
        ribd.run();
    }

    private RefinedMain() {
        // private constructor to prevent instantiation
    }

    /*
     * Checks that certain parameters are consistent, and prints error
     * message and exits if parameters are inconsistent.
     *
     * @param args the command line arguments.
     */
    private static RefinedPar parameters(String[] args) {
        RefinedPar par = new RefinedPar(args);
        checkOutputPrefix(par);
        if (par.window() < 3*par.length()) {
            String s = SHORT_HELP + Const.nl
                    + Const.nl + "ERROR: The \"window\" parameter must be at least "
                    + "3 times the \"ibd\" parameter"
                    + Const.nl + "Exiting program.";
            Utilities.exit(s);
        }
        return par;
    }

    private static void checkOutputPrefix(RefinedPar par) {
        File outPrefix = new File(par.out());
        if (outPrefix.isDirectory()) {
            String s = "ERROR: \"out\" parameter cannot be a directory: \""
                    + par.out() + "\"";
            Utilities.exit(RefinedPar.usage() + s);
        }
    }
}
